/**
  MatchFree is a simple match-3 game.

  Author: Timo Wiren
  Modified: 2014-06-26
 */
using System;
using System.IO;
using System.Diagnostics;
using System.Collections.Generic;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;
using OpenTK.Audio;
using OpenTK.Audio.OpenAL;
using OpenTK.Input;

class Game : GameWindow
{
    public static Vector2 screenSize = new Vector2( 700, 650 );
    public static float ScaleFactor { get; private set; }
    private static string workingDirectory;
    private GameState gameState = new GameState(9, 9, 5);
    private GridGeometry gridGeometry = new GridGeometry();
    private FontRenderer fontRenderer = new FontRenderer();
    private bool hasAudioDevice = false;
    private int soundMoveSrc = 0;
    private long lastActionMs = 0;
    private long lastGravityMs = 0;
    private long gameStartupMs = 0;
    private List< GridGeometry.Puff > puffs = new List< GridGeometry.Puff >();
    private Vector4 gridOffset = new Vector4(-200, -200, 0, 0);
    private int highScore;
    private string highScorePath;

    public Game() :
        base((int)screenSize.X, (int)screenSize.Y, GraphicsMode.Default, "MatchFree" )
    //base( 1024, 768, GraphicsMode.Default, "Game", 0, DisplayDevice.Default, 3, 2, GraphicsContextFlags.ForwardCompatible )
    {
        WindowBorder = WindowBorder.Fixed;
        VSync = VSyncMode.On;
        Mouse.ButtonUp += MouseEventHandler;
        InitAudio();
        MatchFree.Utils.stopWatch.Start();
    }


    private void ShowStartupHint()
    {
        const float startupHintDisplayTimeMs = 8000.0f;

        if (MatchFree.Utils.stopWatch.ElapsedMilliseconds - gameStartupMs < startupHintDisplayTimeMs)
        {
            float opacity = 1 - (MatchFree.Utils.stopWatch.ElapsedMilliseconds - gameStartupMs) / startupHintDisplayTimeMs;
            gridGeometry.shader.SetFloat(opacity, Shader.Uniform.Opacity);
            GL.Enable(EnableCap.Blend);
            GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);
            fontRenderer.DrawText("Swap nearby tiles to get 3 or more in line", -320, -180, 1);
            GL.Disable(EnableCap.Blend);
            gridGeometry.shader.SetVector(Vector2.Zero, Shader.Uniform.Translation);
            gridGeometry.shader.SetFloat(1.0f, Shader.Uniform.Opacity);
        }
    }

    private void InitAudio()
    {
        // OpenAL using C#/OpenTK works on my machine only on Linux and OS X, not on Windows, so don't enable audio on Windows.
        if (Environment.OSVersion.Platform.ToString() != "Unix")
        {
            return;
        }

        try
        {
            AudioContext ac = new AudioContext();
            ac.CheckErrors();
            hasAudioDevice = true;
        }
        catch (AudioException e)
        {
            Console.WriteLine("Unable to init audio device: " + e.ToString());
            return;
        }

        int buffer = AL.GenBuffer();
        soundMoveSrc = AL.GenSource();

        WaveData waveFile = new WaveData(workingDirectory + "/Resources/failed.wav");

        AL.BufferData(buffer, waveFile.SoundFormat, waveFile.SoundData, waveFile.SoundData.Length, waveFile.SampleRate);
        waveFile.dispose();

        AL.Source(soundMoveSrc, ALSourcei.Buffer, buffer);
        AL.Source(soundMoveSrc, ALSourceb.Looping, false);
        AL.GenSources(soundMoveSrc);

        buffer = AL.GenBuffer();

        waveFile = new WaveData(workingDirectory + "/Resources/badmove.wav");

        AL.BufferData(buffer, waveFile.SoundFormat, waveFile.SoundData, waveFile.SoundData.Length, waveFile.SampleRate);
        waveFile.dispose();
    }


    private void MouseEventHandler( object sender, MouseButtonEventArgs buttonEvent )
    {
        if (gameState.movesLeft == 0)
        {
            return;
        }

        if (IsApplyingGravity())
        {
            return;
        }

        int row, column;

        if (gameState.selectedBlockX1 == GameState.noSelection || gameState.selectedBlockY1 == GameState.noSelection)
        {
            gridGeometry.GetBlockCoordinate(buttonEvent.X, buttonEvent.Y, out row, out column);

            if (gameState.CanSelect(row, column))
            {
                gameState.selectedBlockX1 = column;
                gameState.selectedBlockY1 = row;

                if (hasAudioDevice)
                {
                    AL.SourcePlay(soundMoveSrc);
                }
            }

            gridGeometry.CreateSelection1(row, column);
        }
        else if (gameState.selectedBlockX2 == GameState.noSelection || gameState.selectedBlockY2 == GameState.noSelection)
        {
            gridGeometry.GetBlockCoordinate(buttonEvent.X, buttonEvent.Y, out row, out column);

            if (gameState.CanSelect(row, column))
            {
                gameState.selectedBlockX2 = column;
                gameState.selectedBlockY2 = row;

                gridGeometry.CreateSelection1(GameState.noSelection, GameState.noSelection);

                if (gameState.SwappedMatch3())
                {
                    gameState.SwapSelectedBlocks();
                    gameState.DestroyMatches(puffs, MatchFree.Utils.stopWatch.ElapsedMilliseconds);

                    gridGeometry.Create(gameState.GetGrid(), 32, 5, gridOffset);

                    if (hasAudioDevice)
                    {
                        AL.SourcePlay(soundMoveSrc);
                    }
                }
            }

            gameState.selectedBlockX1 = GameState.noSelection;
            gameState.selectedBlockY1 = GameState.noSelection;
            gameState.selectedBlockX2 = GameState.noSelection;
            gameState.selectedBlockY2 = GameState.noSelection;
        }
    }

    protected override void OnLoad( EventArgs e )
    {
        GL.Viewport(0, 0, Width, Height);
		
        LoadAssets();
        gridGeometry.Create(gameState.GetGrid(), 32, 5, gridOffset);
        gridGeometry.InitSelectionGraphicsObjects();
    }


    protected override void OnUnload( EventArgs e )
    {
        if (gameState.score > highScore)
        {
            System.IO.File.WriteAllText(highScorePath, gameState.score.ToString());
        }
    }

    protected override void OnResize( EventArgs e )
    {
        base.OnResize(e);
    }


    protected override void OnUpdateFrame( FrameEventArgs e )
    {
        if (Keyboard[Key.Escape])
        {
            Exit();
        }

        if (MatchFree.Utils.stopWatch.ElapsedMilliseconds - lastActionMs > 500)
        {
            gameState.DestroyMatches(puffs, MatchFree.Utils.stopWatch.ElapsedMilliseconds);

            bool moved = gameState.ApplyGravity();

            if (moved)
            {
                lastGravityMs = MatchFree.Utils.stopWatch.ElapsedMilliseconds;
            }

            gridGeometry.Create(gameState.GetGrid(), 32, 5, gridOffset);
            lastActionMs = MatchFree.Utils.stopWatch.ElapsedMilliseconds;
        }

        var puffsToRemove = new List< GridGeometry.Puff >();

        foreach (var puff in puffs)
        {
            if (MatchFree.Utils.stopWatch.ElapsedMilliseconds - puff.spawnTimeMs > GridGeometry.Puff.lifeTimeMs)
            {
                puffsToRemove.Add(puff);
            }
        }

        puffsToRemove.Clear();
        CheckForGLErrors();
    }


    private bool IsApplyingGravity()
    {
        return MatchFree.Utils.stopWatch.ElapsedMilliseconds - lastGravityMs < 500;
    }


    protected override void OnRenderFrame( FrameEventArgs e )
    {
        GL.ClearColor(IsApplyingGravity() ? new Color4(0.2f, 0.2f, 0.2f, 1.0f) : Color4.Black);
        GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
        gridGeometry.Draw();

        if (gameState.selectedBlockX1 != GameState.noSelection)
        {
            gridGeometry.DrawSelection1(gameState.selectedBlockY1, gameState.selectedBlockX1);
        }

        fontRenderer.DrawText("Moves: " + gameState.movesLeft, 0, 235, 1);
        fontRenderer.DrawText("Score: " + gameState.score, 0, 260, 1);
        fontRenderer.DrawText("High Score: " + highScore, 0, 285, 1);
        gridGeometry.DrawPuffs(puffs);
        ShowStartupHint();

        if (gameState.movesLeft == 0)
        {
            fontRenderer.DrawText("Game Over", -120, -200, 1);

            if (gameState.score > highScore)
            {
                fontRenderer.DrawText("You got high score!", -120, -175, 1);
            }
        }

        SwapBuffers();
    }


    private void LoadAssets()
    {
        Matrix4 orthoMatrix = Matrix4.CreateOrthographic(Width, Height, 0, 1);
        gridGeometry.shader.Load(workingDirectory + "/Resources/shader.vert", workingDirectory + "/Resources/shader.frag");
        gridGeometry.shader.Use();
        gridGeometry.shader.SetMatrix(ref orthoMatrix, Shader.Uniform.ProjectionMatrix);
        fontRenderer.Init(workingDirectory);

        highScorePath = workingDirectory + "/Resources/highscore.txt";

        if (File.Exists(highScorePath))
        {
            int.TryParse(System.IO.File.ReadAllText(highScorePath), out highScore);
        }

        try
        {
            gridGeometry.textureId = MatchFree.Utils.LoadTexture(workingDirectory + "/Resources/blocks.png");
            gridGeometry.selectionTextureId = MatchFree.Utils.LoadTexture(workingDirectory + "/Resources/blocks.png");
            gridGeometry.puffTextureId = MatchFree.Utils.LoadTexture(workingDirectory + "/Resources/puff.png");
        }
        catch (ArgumentException e)
        {
            Console.WriteLine(e.StackTrace);
        }
    }


    private void CheckForGLErrors()
    {
        ErrorCode errorCode;

        while ((errorCode = GL.GetError()) != ErrorCode.NoError)
        {
            Console.WriteLine(errorCode.ToString());
        }
    }


    [STAThread]
    static void Main()
    {
        workingDirectory = new FileInfo(System.Reflection.Assembly.GetExecutingAssembly().Location).DirectoryName;

        Game game = new Game();
        string version = GL.GetString(StringName.Version);
        Console.WriteLine("OpenGL context: " + version);

        ScaleFactor = game.Width / (float)screenSize.X; // derive current DPI scale factor

        // Adjusts scaling for Retina display.
        if (ScaleFactor > 1.999f)
        {
            game.Dispose();
            screenSize /= 2.0f;
            game = new Game();
        }

        game.CursorVisible = true;
        game.Run(30.0);
    }
}
