using System;
using System.Diagnostics;
using System.Collections.Generic;

public class GameState
{
	public static readonly int blockTypeCount = 5;
	public static readonly int noSelection = -1;
	public int selectedBlockX1 = noSelection;
	public int selectedBlockY1 = noSelection;
	public int selectedBlockX2 = noSelection;
	public int selectedBlockY2 = noSelection;
    public const int uninitializedValue = 4;
	private int[,] grid;
    public int movesLeft { get; private set; }
    public int score { get; private set; }
    private Random random = new Random();
   

    public GameState( uint rows, uint columns, int moves )
	{
		grid = new int[ rows, columns ];
        this.movesLeft = moves;
		Generate();
	}


    // Returns true if the grid was modified.
    public bool ApplyGravity()
    {
        bool somethingWasMoved = false;

        for (int col = 0; col < grid.GetLength(1); ++col)
        {
            for (int row = 0; row < grid.GetLength(0); ++row)
            {
                if (grid[row, col] == uninitializedValue)
                {
                    if (row == grid.GetLength(0) - 1)
                    {
                        grid[row, col] = random.Next(blockTypeCount - 1);
                        somethingWasMoved = true;
                    }
                    else
                    {
                        grid[row, col] = grid[row + 1, col];
                        grid[row + 1, col] = uninitializedValue;
                        somethingWasMoved = true;
                    }
                }
            }
        }

        return somethingWasMoved;
    }


    public bool ContainsMatches()
    {
        for (int row = 0; row < grid.GetLength(0); ++row)
        {
            for (int col = 0; col < grid.GetLength(1); ++col)
            {
                if (!CanInit( grid[ row, col ], row, col ))
                {
                    return true;
                }
            }
        }

        return false;
    }


    public bool DestroyMatches( List< GridGeometry.Puff > destroyedLocations, long currentTimeMs )
    {
        bool somethingWasDestroyed = false;

        // Horizontal.
        for (int row = 0; row < grid.GetLength(0); ++row)
        {
            int horizMatches = 1;

            for (int col = 0; col < grid.GetLength(1); ++col)
            {
                if (col > 0 && grid[ row, col ] == grid[ row, col - 1 ])
                {
                    ++horizMatches;
                }
                else
                {
                    if (horizMatches >= 3)
                    {
                        for (int i = 0; i < horizMatches; ++i)
                        {
                            grid[ row, col - i - 1 ] = uninitializedValue;
                            destroyedLocations.Add( new GridGeometry.Puff( row, col - i - 1, currentTimeMs ) );
                        }

                        score += horizMatches * 10;

                        somethingWasDestroyed = true;
                    }

                    horizMatches = 1;
                }
            }

            if (horizMatches >= 3)
            {
                for (int i = 0; i < horizMatches; ++i)
                {
                    grid[ row, grid.GetLength(1) - i - 1 ] = uninitializedValue;
                    destroyedLocations.Add( new GridGeometry.Puff( row, grid.GetLength(1) - i - 1, currentTimeMs ) );
                }

                score += horizMatches * 10;
                somethingWasDestroyed = true;
            }
        }

        // Vertical.
        for (int col = 0; col < grid.GetLength(1); ++col)
        {
            int vertMatches = 1;

            for (int row = 0; row < grid.GetLength(0); ++row)
            {
                if (row > 0 && grid[ row, col ] == grid[ row - 1, col ])
                {
                    ++vertMatches;
                }
                else
                {
                    if (vertMatches >= 3)
                    {
                        for (int i = 0; i < vertMatches; ++i)
                        {
                            grid[ row - i - 1, col ] = uninitializedValue;
                            destroyedLocations.Add( new GridGeometry.Puff( row - i - 1, col, currentTimeMs ) );
                        }

                        somethingWasDestroyed = true;
                        score += vertMatches * 10;
                    }
                        
                    vertMatches = 1;
                }
            }

            if (vertMatches >= 3)
            {
                for (int i = 0; i < vertMatches; ++i)
                {
                    grid[ grid.GetLength(0) - i - 1, col ] = uninitializedValue;
                    destroyedLocations.Add( new GridGeometry.Puff( grid.GetLength(0) - i - 1, col, currentTimeMs ) );
                }

                score += vertMatches * 10;
                somethingWasDestroyed = true;
            }
        }

        return somethingWasDestroyed;
    }


	public void Generate()
	{
		for (int row = 0; row < grid.GetLength( 0 ); ++row)
		{
			for (int col = 0; col < grid.GetLength( 1 ); ++col)
			{
				grid[ row, col ] = uninitializedValue;

                // Tries to find a non-match at most 10 times and restarts if fails.
				for (int i = 0; i < 10; ++i)
				{
                    int candidate = random.Next( blockTypeCount-1 );

					if (CanInit(candidate, row, col))
					{
						grid[ row, col ] = candidate;
						break;
					}
				}

				if (grid[ row, col ] == uninitializedValue)
				{
					Generate();
				}
			}
		}
	}


	public bool CanSelect( int row, int column )
	{
        bool selectionInGrid = row > -1 && row < grid.GetLength(0) && column > -1 && column < grid.GetLength(1);

        if (selectedBlockX1 == noSelection && selectionInGrid)
		{
			return true;
		}

		// Selection must be next to an already selected block.
        if (row > -1 && selectedBlockY1 - 1 == row && selectedBlockX1 == column)
		{
			return true;
		}
        else if (row > -1 && selectedBlockY1 + 1 == row && selectedBlockX1 == column)
		{
			return true;
		}
        else if (column > -1 && selectedBlockX1 + 1 == column && selectedBlockY1 == row)
		{
			return true;
		}
        else if (column > -1 && selectedBlockX1 - 1 == column && selectedBlockY1 == row)
		{
			return true;
		}

        Console.WriteLine( string.Format("Cannot select row {0}, col {1}", row, column ) );
		return false;
	}


	public void SwapSelectedBlocks()
	{
		Debug.Assert( selectedBlockX1 != noSelection && selectedBlockX2 != noSelection &&
			          selectedBlockY1 != noSelection && selectedBlockY2 != noSelection,
			          "Both selections must be active.");

		int originalSelection1 = grid[ selectedBlockY1, selectedBlockX1 ];
		grid[ selectedBlockY1, selectedBlockX1 ] = grid[ selectedBlockY2, selectedBlockX2 ];
		grid[ selectedBlockY2, selectedBlockX2 ] = originalSelection1;
	}


	public int[,] GetGrid()
	{
		return grid;
	}


	public void Print()
	{
		for (int row = 0; row < grid.GetLength( 0 ); ++row)
		{
			for (int col = 0; col < grid.GetLength( 1 ); ++col)
			{
				Console.Write( grid[ row, col ] );
			}

			Console.WriteLine();
		}
	}


	/// <summary>
    /// If swapping selected tiles causes a match-3, returns true. Decreases moves if successful.
	/// </summary>
	public bool SwappedMatch3()
	{
		SwapSelectedBlocks();

		if (!CanInit( grid[ selectedBlockY2, selectedBlockX2 ], selectedBlockY2, selectedBlockX2 ) ||
			!CanInit( grid[ selectedBlockY1, selectedBlockX1 ], selectedBlockY1, selectedBlockX1 ))
		{
			SwapSelectedBlocks();
            --movesLeft;

			return true;
		}

		SwapSelectedBlocks();
		return false;
	}


	/// <summary>
	/// Returns true if value can be inserted into grid[ row, column ] without causing a 3-match.
	/// </summary>
	private bool CanInit( int value, int row, int column )
	{
		int matchHorizontalCount = 0;

		// Left 1
        bool left1ok = column - 1 > -1 && grid[row, column - 1] == value;
        if (left1ok)
		{
            //Console.WriteLine( "left 1 ok" );
			++matchHorizontalCount;
		}
        else if (column - 1 > -1)
        {
            //Console.WriteLine( "left 1 NOT ok, color is " + colors[ grid[row, column - 1] ] );
        }

		// Left 2
        if (left1ok && column - 2 > -1 && grid[ row, column - 2 ] == value)
		{
            //Console.WriteLine( "left 2 ok" );
			++matchHorizontalCount;
		}
        else if (column - 2 > -1)
        {
            //Console.WriteLine( "left 2 NOT ok, color is " + colors[ grid[row, column - 2] ] );
        }

		// Right 1
        bool right1ok = column + 1 < grid.GetLength(1) && grid[row, column + 1] == value;
        if (right1ok)
		{
            //Console.WriteLine( "right 1 ok" );
			++matchHorizontalCount;
		}
        else if (column + 1 < grid.GetLength(1))
        {
            //Console.WriteLine( "right 1 NOT ok, color is " + colors[ grid[row, column + 1] ] );
        }

		// Right 2
        if (right1ok && column + 2 < grid.GetLength( 1 ) && grid[ row, column + 2 ] == value)
		{
            //Console.WriteLine( "right 2 ok" );

			++matchHorizontalCount;
		}
        else if (column + 2 < grid.GetLength(1))
        {
            //Console.WriteLine( "right 2 NOT ok, color is " + colors[ grid[row, column + 2] ] );
        }
            
		if (matchHorizontalCount >= 2)
		{
			return false;
		}

		int matchVerticalCount = 0;

        // Down 1
        bool down1ok = row - 1 > -1 && grid[row - 1, column] == value;
            
        if (down1ok)
        {
            //Console.WriteLine("down 1 ok");

            ++matchVerticalCount;
        }
        else if (row - 1 > -1)
        {
            //Console.WriteLine( "down 1 NOT ok, color is " + colors[grid[row - 1, column]]);
        }

        // Down 2
        if (down1ok && row - 2 > -1 && grid[ row - 2, column ] == value)
		{
            //Console.WriteLine( "down 2 ok" );

			++matchVerticalCount;
		}
        else if (row - 2 > -1)
        {
            //Console.WriteLine( "down 2 NOT ok, color is " + colors[grid[row - 2, column]]);
        }

        // Up 1
        bool up1ok = row + 1 < grid.GetLength(0) - 1 && grid[row + 1, column] == value;

        if (up1ok)
		{
            // Console.WriteLine( "up 1 ok" );

			++matchVerticalCount;
		}
        else if (row + 1 < grid.GetLength(0) - 1)
        {
            //Console.WriteLine( "up1 NOT ok, color is " + colors[ grid[ row + 1, column ] ] );
        }

        // Up 2
        if (up1ok && row + 2 < grid.GetLength(0) && grid[row + 2, column] == value)
        {
            //Console.WriteLine("up 2 ok");

            ++matchVerticalCount;
        }

		return matchVerticalCount < 2;
	}
}

